"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


class CheerEvent(SpecificBroadcaster):
    """
    Represents a channel cheer event on Twitch.

    Attributes
    ----------
    is_anonymous: bool
        Indicates whether the cheer was made anonymously.
    user_id: Optional[str]
        The ID of the user who made the cheer, if not anonymous.
    user_login: Optional[str]
        The login name of the user who made the cheer, if not anonymous.
    user_name: Optional[str]
        The display name of the user who made the cheer, if not anonymous.
    message: str
        The message accompanying the cheer.
    bits: int
        The number of bits cheered.
    """
    is_anonymous: bool
    user_id: Optional[str]
    user_login: Optional[str]
    user_name: Optional[str]
    message: str
    bits: int
