"""
twitch.ext.overlay

An extension that provides functionalities to manage and send overlay alerts,
handle WebSocket connections, and manage file attachments.

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""


from .overlay import Overlay
